# poppler-python: python binding to the poppler-cpp pdf lib
# Copyright (C) 2020, Charles Brunet <charles@cbrunet.net>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import pytest

from poppler.rectangle import Rectangle


def test_init_from_int():
    x, y, w, h = 10, 12, 30, 40
    rectangle = Rectangle(x, y, w, h)

    assert rectangle.x == x
    assert rectangle.y == y
    assert rectangle.width == w
    assert rectangle.height == h


def test_init_from_float():
    x, y, w, h = 5.5, 10.1, 20.2, 30.3
    rectangle = Rectangle(x, y, w, h)

    assert pytest.approx(rectangle.x, 0.0001) == x
    assert pytest.approx(rectangle.y, 0.0001) == y
    assert pytest.approx(rectangle.width, 0.0001) == w
    assert pytest.approx(rectangle.height, 0.0001) == h


def test_init_default():
    rectangle = Rectangle()
    assert rectangle.is_empty()


def test_getters():
    rectangle = Rectangle(10, 20, 30, 40)
    assert rectangle.left == 10
    assert rectangle.right == 40
    assert rectangle.top == 20
    assert rectangle.bottom == 60


def test_setters():
    rectangle = Rectangle()
    rectangle.left = 10
    rectangle.right = 20
    rectangle.top = 30
    rectangle.bottom = 50

    assert rectangle.x == 10
    assert rectangle.y == 30
    assert rectangle.width == 10
    assert rectangle.height == 20


def test_rect_as_tuple():
    rectangle = Rectangle(10, 20, 30, 40)
    t = rectangle.as_tuple()
    assert t == (10, 20, 40, 60)



# poppler-python: python binding to the poppler-cpp pdf lib
# Copyright (C) 2020, Charles Brunet <charles@cbrunet.net>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import pytest

from poppler import TransitionType as Type
from poppler import Alignment, Direction


@pytest.mark.parametrize(
    "page_num,transition_type,alignment,direction",
    [
        (3, Type.blinds, Alignment.horizontal, Direction.inward),
        (4, Type.blinds, Alignment.vertical, Direction.inward),
        (5, Type.box, Alignment.horizontal, Direction.inward),
        (6, Type.box, Alignment.horizontal, Direction.outward),
        (7, Type.cover, Alignment.horizontal, Direction.inward),
        (8, Type.dissolve, Alignment.horizontal, Direction.inward),
        (9, Type.fade, Alignment.horizontal, Direction.inward),
        (10, Type.glitter, Alignment.horizontal, Direction.inward),
        (11, Type.push, Alignment.horizontal, Direction.inward),
        (12, Type.replace, Alignment.horizontal, Direction.inward),
        (13, Type.split, Alignment.vertical, Direction.inward),
        (14, Type.split, Alignment.vertical, Direction.outward),
        (15, Type.split, Alignment.horizontal, Direction.inward),
        (16, Type.split, Alignment.horizontal, Direction.outward),
        (17, Type.wipe, Alignment.horizontal, Direction.inward),
        (18, Type.uncover, Alignment.horizontal, Direction.inward),
    ],
)
def test_transition_type(
    sample_document, page_num, transition_type, alignment, direction
):
    page = sample_document.create_page(page_num)
    transition = page.transition()

    assert transition is not None
    assert transition.type == transition_type
    assert transition.alignment == alignment
    assert transition.direction == direction
    assert transition.is_rectangular is False
    assert transition.scale == 1.0


@pytest.mark.parametrize(
    "page_num,angle", [(7, 0), (10, 315), (11, 90), (17, 180), (18, 270)]
)
def test_transition_angle(sample_document, page_num, angle):
    page = sample_document.create_page(page_num)
    transition = page.transition()

    assert transition.angle == angle


@pytest.mark.parametrize("page_num,duration", [(3, 0), (4, 1), (5, 1), (6, 2), (7, 15)])
def test_transition_duration(sample_document, page_num, duration):
    page = sample_document.create_page(page_num)
    transition = page.transition()

    assert transition.duration == duration



# poppler-python: python binding to the poppler-cpp pdf lib
# Copyright (C) 2020, Charles Brunet <charles@cbrunet.net>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


"""This module contains utility functions that are not part of
poppler-cpp, but that helps interacting with it.
"""

from datetime import datetime
import functools

from poppler.cpp.version import version_major, version_minor, version_micro


def from_time_type(timestamp):
    """Convert a timestamp to a :class:`datetime.datetime` object.

    Args:
        timestamp (int): a positive integer, to be interpreted as 32 bits complement 2.

    Returns:
        Optional[datetime.datetime]: the date and time,
                                     or None if the timestamp is equivalent to -1.

    """
    if timestamp == 2 ** 32 - 1:
        return None
    return datetime.fromtimestamp(timestamp)


def to_time_type(date_time):
    """Convert a Python :class:`datetime.datetime` object
    to a timestamp compatible with poppler.

    Args:
        date_time (Optional[datetime.datetime]): the date and time, or None

    Returns:
        int: the timestamp

    """
    return int(date_time.timestamp()) if date_time else 2 ** 32 - 1


def version():
    """Get poppler version, as a tuple

    Returns:
        Tuple[int, int, int]: (major, minor, micro)

    """
    return version_major(), version_minor(), version_micro()


def since(major, minor):
    """Decorator used to mark the minimum required version of poppler
    needed to execute a function.

    If poppler version is lower than the specified version,
    it raises a NotImplementedError when called.

    Args:
        major (int): major version number
        minor (int): minor version number

    """
    def wrapper(fct):
        @functools.wraps(fct)
        def wrapped(*args, **kwargs):
            if version() < (major, minor, 0):
                raise NotImplementedError(
                    "This functionality requires at least Poppler version {}".format(
                        ".".join(map(str, (major, minor, 0)))
                    )
                )
            return fct(*args, **kwargs)

        return wrapped

    return wrapper



# poppler-python: python binding to the poppler-cpp pdf lib
# Copyright (C) 2020, Charles Brunet <charles@cbrunet.net>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from poppler.cpp import document
from poppler.cpp.global_ import ustring
from poppler.utilities import from_time_type, to_time_type
from poppler.destination import Destination
from poppler.embeddedfile import EmbeddedFile
from poppler.font import FontInfo, FontIterator
from poppler.page import Page
from poppler.toc import Toc
from poppler.utilities import since

from collections import namedtuple
from functools import singledispatch, wraps
from pathlib import Path


PDFId = namedtuple("PDFId", ["permanent_id", "update_id"])


class LockedDocumentError(RuntimeError):
    def __init__(self):
        super().__init__("Cannot perform this operation on a locked document.")


def ensure_unlocked(fct):
    @wraps(fct)
    def wrapper(self, *args, **kwargs):
        if self._document.is_locked():
            raise LockedDocumentError()
        return fct(self, *args, **kwargs)

    return wrapper


class Document:

    PageLayout = document.page_layout_enum
    PageMode = document.page_mode_enum

    def __init__(self, poppler_document, data=None):
        if poppler_document is None:
            raise ValueError("Corrupted or invalid PDF document")

        self._document = poppler_document
        self._data = data

    @ensure_unlocked
    def create_font_iterator(self, page=0):
        return FontIterator(self._document.create_font_iterator(page))

    @ensure_unlocked
    def create_toc(self):
        t = self._document.create_toc()
        return Toc(t) if t else None

    @ensure_unlocked
    def create_page(self, index):
        return Page(self._document.create_page(index))

    @property
    @ensure_unlocked
    def author(self):
        return self.info_key("Author")

    @author.setter
    @since(0, 46)
    @ensure_unlocked
    def author(self, author):
        self.set_info_key("Author", author)

    @property
    @ensure_unlocked
    def creation_date(self):
        return self.info_date("CreationDate")

    @creation_date.setter
    @since(0, 46)
    @ensure_unlocked
    def creation_date(self, creation_date):
        self.set_info_date("CreationDate", creation_date)

    @property
    @ensure_unlocked
    def creator(self):
        return self.info_key("Creator")

    @creator.setter
    @since(0, 46)
    @ensure_unlocked
    def creator(self, creator):
        self.set_info_key("Creator", creator)

    @property
    @ensure_unlocked
    def keywords(self):
        return self.info_key("Keywords")

    @keywords.setter
    @since(0, 46)
    @ensure_unlocked
    def keywords(self, keywords):
        self.set_info_key("Keywords", keywords)

    @property
    @ensure_unlocked
    def metadata(self):
        meta = self._document.metadata()
        return str(meta)

    @property
    @ensure_unlocked
    def modification_date(self):
        return self.info_date("ModDate")

    @modification_date.setter
    @since(0, 46)
    @ensure_unlocked
    def modification_date(self, modification_date):
        self.set_info_date("ModDate", modification_date)

    @property
    def pdf_id(self):
        return PDFId(*self._document.get_pdf_id())

    @property
    @ensure_unlocked
    def pdf_version(self):
        return self._document.get_pdf_version()

    @property
    @ensure_unlocked
    def producer(self):
        return self.info_key("Producer")

    @producer.setter
    @since(0, 46)
    @ensure_unlocked
    def producer(self, producer):
        self.set_info_key("Producer", producer)

    @property
    @ensure_unlocked
    def subject(self):
        return self.info_key("Subject")

    @subject.setter
    @since(0, 46)
    @ensure_unlocked
    def subject(self, subject):
        self.set_info_key("Subject", subject)

    @property
    @ensure_unlocked
    def title(self):
        return self.info_key("Title")

    @title.setter
    @since(0, 46)
    @ensure_unlocked
    def title(self, title):
        self.set_info_key("Title", title)

    @property
    @ensure_unlocked
    def pages(self):
        return self._document.pages()

    @since(0, 74)
    @ensure_unlocked
    def create_destination_map(self):
        return {
            name: Destination(destination)
            for name, destination in self._document.create_destination_map().items()
        }

    @ensure_unlocked
    def embedded_files(self):
        return [EmbeddedFile(f) for f in self._document.embedded_files()]

    @ensure_unlocked
    def fonts(self):
        return [FontInfo(i) for i in self._document.fonts()]

    @ensure_unlocked
    def has_embedded_files(self):
        return self._document.has_embedded_files()

    def has_permission(self, which):
        return self._document.has_permission(which)

    @ensure_unlocked
    def info_date(self, key):
        timestamp = self._document.info_date(key)
        return from_time_type(timestamp)

    @since(0, 46)
    @ensure_unlocked
    def set_info_date(self, key, val):
        return self._document.set_info_date(key, to_time_type(val))

    @ensure_unlocked
    def info_key(self, key):
        info = self._document.info_key(key)
        return str(info)

    @ensure_unlocked
    def infos(self):
        """Get the document info dictionary as a dict object."""
        info_dict = {}
        for key in self._document.info_keys():
            if key.endswith("Date"):
                info_dict[key] = self.info_date(key)
            else:
                info_dict[key] = self.info_key(key)
        return info_dict

    @since(0, 46)
    @ensure_unlocked
    def set_info_key(self, key, val):
        return self._document.set_info_key(key, ustring(val))

    @ensure_unlocked
    def info_keys(self):
        return self._document.info_keys()

    def is_encrypted(self):
        return self._document.is_encrypted()

    def is_linearized(self):
        return self._document.is_linearized()

    def is_locked(self):
        return self._document.is_locked()

    @property
    @ensure_unlocked
    def page_layout(self):
        return self._document.page_layout()

    @property
    @ensure_unlocked
    def page_mode(self):
        return self._document.page_mode()

    @since(0, 46)
    @ensure_unlocked
    def remove_info(self):
        return self._document.remove_info()

    @since(0, 46)
    @ensure_unlocked
    def save(self, file_name):
        return self._document.save(str(file_name))

    @since(0, 46)
    @ensure_unlocked
    def save_a_copy(self, file_name):
        return self._document.save_a_copy(str(file_name))

    def unlock(self, owner_password, user_password):
        return self._document.unlock(owner_password, user_password)


def load_from_file(file_name, owner_password=None, user_password=None):
    return Document(
        document.load_from_file(
            str(file_name), owner_password or "", user_password or ""
        )
    )


def load_from_data(file_data: bytes, owner_password=None, user_password=None):
    return Document(
        document.load_from_data(file_data, owner_password or "", user_password or ""),
        file_data
    )


@singledispatch
def load(arg, owner_password=None, user_password=None):
    try:
        data = arg.read()
        return load_from_data(data, owner_password, user_password)

    except AttributeError:
        raise TypeError(
            "Load cannot be called with argument of type {}".format(type(arg))
        )

    except UnicodeDecodeError:
        raise TypeError("Stream must be read as bytes.")


@load.register(str)
def _(arg: str, owner_password=None, user_password=None):
    return load_from_file(arg, owner_password, user_password)


@load.register(Path)
def _(arg: Path, owner_password=None, user_password=None):
    return load_from_file(arg, owner_password, user_password)


@load.register(bytes)
def _(arg: bytes, owner_password=None, user_password=None):
    return load_from_data(arg, owner_password, user_password)



# poppler-python: python binding to the poppler-cpp pdf lib
# Copyright (C) 2020, Charles Brunet <charles@cbrunet.net>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import pytest

from poppler.page import Page
from poppler.rectangle import Rectangle
from poppler import version
from poppler import CaseSensitivity


def test_page_duration(pdf_page):
    assert pdf_page.duration == -1.0


@pytest.mark.skipif(version() < (0, 46, 0), reason="Requires at least Poppler 0.46.0")
def test_page_label(pdf_page):
    assert pdf_page.label == "1"


def test_page_orientation(pdf_page):
    assert pdf_page.orientation == Page.Orientation.portrait


def test_page_rect(pdf_page):
    r = pdf_page.page_rect()
    assert r.as_tuple() == (0.0, 0.0, 612.0, 792.0)


@pytest.mark.parametrize(
    "box",
    [
        Page.PageBox.media_box,
        Page.PageBox.crop_box,
        Page.PageBox.bleed_box,
        Page.PageBox.trim_box,
        Page.PageBox.art_box,
    ],
)
def test_page_rect_box(pdf_page, box):
    r = pdf_page.page_rect(box)
    assert r.as_tuple() == (0.0, 0.0, 612.0, 792.0)


def test_text(pdf_page):
    text = pdf_page.text()
    if version() < (0, 46, 0):
        expected = "Page "
    elif version() < (0, 88, 0):
        expected = "Page 1"
    else:
        expected = "Page 1\n\x0c"
    assert text == expected


@pytest.mark.skipif(version() < (0, 63, 0), reason="Requires at least Poppler 0.63.0")
def test_text_list(pdf_page):
    text_list = pdf_page.text_list()
    assert len(text_list) == 2

    text_box = text_list[0]
    assert text_box.text == "Page"
    assert pytest.approx(text_box.bbox.as_tuple(), abs=0.1) == (56.8, 57.2, 80.1, 70.5)
    if version() >= (0, 68, 0):
        assert text_box.rotation == 0
    assert pytest.approx(text_box.char_bbox(0).as_tuple(), abs=0.1) == (
        56.8,
        57.2,
        63.5,
        70.5,
    )
    assert text_box.has_space_after is True


@pytest.mark.xfail(version() <= (0, 62, 0), reason="Is it a bug in Poppler 0.62?")
def test_search_found(pdf_page):
    r = Rectangle(0.0, 0.0, 0.0, 0.0)
    result = pdf_page.search(
        "Page", r, Page.SearchDirection.from_top, CaseSensitivity.case_sensitive
    )

    assert isinstance(result, Rectangle)
    assert pytest.approx(result.as_tuple(), abs=0.1) == (56.8, 57.2, 80.1, 70.5)


def test_search_not_found(pdf_page):
    r = Rectangle(0.0, 0.0, 612.0, 792.0)
    result = pdf_page.search(
        "Stchroumph", r, Page.SearchDirection.from_top, CaseSensitivity.case_sensitive
    )

    assert result is None


def test_page_transition(pdf_page):
    assert pdf_page.transition() is None


@pytest.mark.skipif(version() < (0, 89, 0), reason="Requires at least Poppler 0.89.0")
def test_text_box_font_info(pdf_page):
    boxes = pdf_page.text_list(pdf_page.TextListOption.text_list_include_font)
    box = boxes[0]

    assert box.has_font_info
    assert box.get_font_name().endswith("LiberationSerif")
    assert pytest.approx(box.get_font_size(), abs=0.1) == 12.0
    assert box.get_wmode() == box.WritingMode.horizontal_wmode



# poppler-python: python binding to the poppler-cpp pdf lib
# Copyright (C) 2020, Charles Brunet <charles@cbrunet.net>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from poppler.image import Image
from poppler import PageRenderer


def test_invalid_image():
    image = Image()
    assert not image.is_valid
    assert image.format == image.Format.invalid


def test_supported_image_formats():
    formats = Image.supported_image_formats()
    assert "png" in formats
    assert "jpeg" in formats


def test_data_size(pdf_page):
    renderer = PageRenderer()
    image = renderer.render_page(pdf_page)
    data = image.data

    assert len(data) == image.bytes_per_row * image.height


def test_image_format_to_str():
    assert str(Image.Format.argb32) == "BGRA"
    assert str(Image.Format.invalid) == ""


def test_image_memory_view(pdf_page):
    renderer = PageRenderer()
    image = renderer.render_page(pdf_page)

    v = image.memoryview()

    assert v.ndim == 3
    assert v.shape == (image.height, image.width, 4)
    assert v.contiguous is True
    assert v.tobytes() == image.data



# poppler-python: python binding to the poppler-cpp pdf lib
# Copyright (C) 2020, Charles Brunet <charles@cbrunet.net>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import pytest

from pathlib import Path

from poppler import load_from_file


@pytest.fixture(scope="session")
def data_path():
    return Path(__file__).parent / "data"


@pytest.fixture()
def pdf_document(data_path):
    return load_from_file(data_path / "document.pdf", "owner", "user")


@pytest.fixture
def pdf_page(pdf_document):
    return pdf_document.create_page(0)


@pytest.fixture(scope="session")
def sample_document(data_path):
    return load_from_file(data_path / "sample.pdf")



# -*- coding: utf-8 -*-
import pytest

import env
from pybind11_tests import IncType, UserType
from pybind11_tests import builtin_casters as m


def test_simple_string():
    assert m.string_roundtrip("const char *") == "const char *"


def test_unicode_conversion():
    """Tests unicode conversion and error reporting."""
    assert m.good_utf8_string() == u"Say utf8‽ 🎂 𝐀"
    assert m.good_utf16_string() == u"b‽🎂𝐀z"
    assert m.good_utf32_string() == u"a𝐀🎂‽z"
    assert m.good_wchar_string() == u"a⸘𝐀z"
    if hasattr(m, "has_u8string"):
        assert m.good_utf8_u8string() == u"Say utf8‽ 🎂 𝐀"

    with pytest.raises(UnicodeDecodeError):
        m.bad_utf8_string()

    with pytest.raises(UnicodeDecodeError):
        m.bad_utf16_string()

    # These are provided only if they actually fail (they don't when 32-bit and under Python 2.7)
    if hasattr(m, "bad_utf32_string"):
        with pytest.raises(UnicodeDecodeError):
            m.bad_utf32_string()
    if hasattr(m, "bad_wchar_string"):
        with pytest.raises(UnicodeDecodeError):
            m.bad_wchar_string()
    if hasattr(m, "has_u8string"):
        with pytest.raises(UnicodeDecodeError):
            m.bad_utf8_u8string()

    assert m.u8_Z() == "Z"
    assert m.u8_eacute() == u"é"
    assert m.u16_ibang() == u"‽"
    assert m.u32_mathbfA() == u"𝐀"
    assert m.wchar_heart() == u"♥"
    if hasattr(m, "has_u8string"):
        assert m.u8_char8_Z() == "Z"


def test_single_char_arguments():
    """Tests failures for passing invalid inputs to char-accepting functions"""

    def toobig_message(r):
        return "Character code point not in range({:#x})".format(r)

    toolong_message = "Expected a character, but multi-character string found"

    assert m.ord_char(u"a") == 0x61  # simple ASCII
    assert m.ord_char_lv(u"b") == 0x62
    assert (
        m.ord_char(u"é") == 0xE9
    )  # requires 2 bytes in utf-8, but can be stuffed in a char
    with pytest.raises(ValueError) as excinfo:
        assert m.ord_char(u"Ā") == 0x100  # requires 2 bytes, doesn't fit in a char
    assert str(excinfo.value) == toobig_message(0x100)
    with pytest.raises(ValueError) as excinfo:
        assert m.ord_char(u"ab")
    assert str(excinfo.value) == toolong_message

    assert m.ord_char16(u"a") == 0x61
    assert m.ord_char16(u"é") == 0xE9
    assert m.ord_char16_lv(u"ê") == 0xEA
    assert m.ord_char16(u"Ā") == 0x100
    assert m.ord_char16(u"‽") == 0x203D
    assert m.ord_char16(u"♥") == 0x2665
    assert m.ord_char16_lv(u"♡") == 0x2661
    with pytest.raises(ValueError) as excinfo:
        assert m.ord_char16(u"🎂") == 0x1F382  # requires surrogate pair
    assert str(excinfo.value) == toobig_message(0x10000)
    with pytest.raises(ValueError) as excinfo:
        assert m.ord_char16(u"aa")
    assert str(excinfo.value) == toolong_message

    assert m.ord_char32(u"a") == 0x61
    assert m.ord_char32(u"é") == 0xE9
    assert m.ord_char32(u"Ā") == 0x100
    assert m.ord_char32(u"‽") == 0x203D
    assert m.ord_char32(u"♥") == 0x2665
    assert m.ord_char32(u"🎂") == 0x1F382
    with pytest.raises(ValueError) as excinfo:
        assert m.ord_char32(u"aa")
    assert str(excinfo.value) == toolong_message

    assert m.ord_wchar(u"a") == 0x61
    assert m.ord_wchar(u"é") == 0xE9
    assert m.ord_wchar(u"Ā") == 0x100
    assert m.ord_wchar(u"‽") == 0x203D
    assert m.ord_wchar(u"♥") == 0x2665
    if m.wchar_size == 2:
        with pytest.raises(ValueError) as excinfo:
            assert m.ord_wchar(u"🎂") == 0x1F382  # requires surrogate pair
        assert str(excinfo.value) == toobig_message(0x10000)
    else:
        assert m.ord_wchar(u"🎂") == 0x1F382
    with pytest.raises(ValueError) as excinfo:
        assert m.ord_wchar(u"aa")
    assert str(excinfo.value) == toolong_message

    if hasattr(m, "has_u8string"):
        assert m.ord_char8(u"a") == 0x61  # simple ASCII
        assert m.ord_char8_lv(u"b") == 0x62
        assert (
            m.ord_char8(u"é") == 0xE9
        )  # requires 2 bytes in utf-8, but can be stuffed in a char
        with pytest.raises(ValueError) as excinfo:
            assert m.ord_char8(u"Ā") == 0x100  # requires 2 bytes, doesn't fit in a char
        assert str(excinfo.value) == toobig_message(0x100)
        with pytest.raises(ValueError) as excinfo:
            assert m.ord_char8(u"ab")
        assert str(excinfo.value) == toolong_message


def test_bytes_to_string():
    """Tests the ability to pass bytes to C++ string-accepting functions.  Note that this is
    one-way: the only way to return bytes to Python is via the pybind11::bytes class."""
    # Issue #816

    def to_bytes(s):
        b = s if env.PY2 else s.encode("utf8")
        assert isinstance(b, bytes)
        return b

    assert m.strlen(to_bytes("hi")) == 2
    assert m.string_length(to_bytes("world")) == 5
    assert m.string_length(to_bytes("a\x00b")) == 3
    assert m.strlen(to_bytes("a\x00b")) == 1  # C-string limitation

    # passing in a utf8 encoded string should work
    assert m.string_length(u"💩".encode("utf8")) == 4


@pytest.mark.skipif(not hasattr(m, "has_string_view"), reason="no <string_view>")
def test_string_view(capture):
    """Tests support for C++17 string_view arguments and return values"""
    assert m.string_view_chars("Hi") == [72, 105]
    assert m.string_view_chars("Hi 🎂") == [72, 105, 32, 0xF0, 0x9F, 0x8E, 0x82]
    assert m.string_view16_chars(u"Hi 🎂") == [72, 105, 32, 0xD83C, 0xDF82]
    assert m.string_view32_chars(u"Hi 🎂") == [72, 105, 32, 127874]
    if hasattr(m, "has_u8string"):
        assert m.string_view8_chars("Hi") == [72, 105]
        assert m.string_view8_chars(u"Hi 🎂") == [72, 105, 32, 0xF0, 0x9F, 0x8E, 0x82]

    assert m.string_view_return() == u"utf8 secret 🎂"
    assert m.string_view16_return() == u"utf16 secret 🎂"
    assert m.string_view32_return() == u"utf32 secret 🎂"
    if hasattr(m, "has_u8string"):
        assert m.string_view8_return() == u"utf8 secret 🎂"

    with capture:
        m.string_view_print("Hi")
        m.string_view_print("utf8 🎂")
        m.string_view16_print(u"utf16 🎂")
        m.string_view32_print(u"utf32 🎂")
    assert (
        capture
        == u"""
        Hi 2
        utf8 🎂 9
        utf16 🎂 8
        utf32 🎂 7
    """
    )
    if hasattr(m, "has_u8string"):
        with capture:
            m.string_view8_print("Hi")
            m.string_view8_print(u"utf8 🎂")
        assert (
            capture
            == u"""
            Hi 2
            utf8 🎂 9
        """
        )

    with capture:
        m.string_view_print("Hi, ascii")
        m.string_view_print("Hi, utf8 🎂")
        m.string_view16_print(u"Hi, utf16 🎂")
        m.string_view32_print(u"Hi, utf32 🎂")
    assert (
        capture
        == u"""
        Hi, ascii 9
        Hi, utf8 🎂 13
        Hi, utf16 🎂 12
        Hi, utf32 🎂 11
    """
    )
    if hasattr(m, "has_u8string"):
        with capture:
            m.string_view8_print("Hi, ascii")
            m.string_view8_print(u"Hi, utf8 🎂")
        assert (
            capture
            == u"""
            Hi, ascii 9
            Hi, utf8 🎂 13
        """
        )

    assert m.string_view_bytes() == b"abc \x80\x80 def"
    assert m.string_view_str() == u"abc ‽ def"
    assert m.string_view_from_bytes(u"abc ‽ def".encode("utf-8")) == u"abc ‽ def"
    if hasattr(m, "has_u8string"):
        assert m.string_view8_str() == u"abc ‽ def"
    if not env.PY2:
        assert m.string_view_memoryview() == "Have some 🎂".encode()

    assert m.bytes_from_type_with_both_operator_string_and_string_view() == b"success"
    assert m.str_from_type_with_both_operator_string_and_string_view() == "success"


def test_integer_casting():
    """Issue #929 - out-of-range integer values shouldn't be accepted"""
    assert m.i32_str(-1) == "-1"
    assert m.i64_str(-1) == "-1"
    assert m.i32_str(2000000000) == "2000000000"
    assert m.u32_str(2000000000) == "2000000000"
    if env.PY2:
        assert m.i32_str(long(-1)) == "-1"  # noqa: F821 undefined name 'long'
        assert m.i64_str(long(-1)) == "-1"  # noqa: F821 undefined name 'long'
        assert (
            m.i64_str(long(-999999999999))  # noqa: F821 undefined name 'long'
            == "-999999999999"
        )
        assert (
            m.u64_str(long(999999999999))  # noqa: F821 undefined name 'long'
            == "999999999999"
        )
    else:
        assert m.i64_str(-999999999999) == "-999999999999"
        assert m.u64_str(999999999999) == "999999999999"

    with pytest.raises(TypeError) as excinfo:
        m.u32_str(-1)
    assert "incompatible function arguments" in str(excinfo.value)
    with pytest.raises(TypeError) as excinfo:
        m.u64_str(-1)
    assert "incompatible function arguments" in str(excinfo.value)
    with pytest.raises(TypeError) as excinfo:
        m.i32_str(-3000000000)
    assert "incompatible function arguments" in str(excinfo.value)
    with pytest.raises(TypeError) as excinfo:
        m.i32_str(3000000000)
    assert "incompatible function arguments" in str(excinfo.value)

    if env.PY2:
        with pytest.raises(TypeError) as excinfo:
            m.u32_str(long(-1))  # noqa: F821 undefined name 'long'
        assert "incompatible function arguments" in str(excinfo.value)
        with pytest.raises(TypeError) as excinfo:
            m.u64_str(long(-1))  # noqa: F821 undefined name 'long'
        assert "incompatible function arguments" in str(excinfo.value)


def test_int_convert():
    class Int(object):
        def __int__(self):
            return 42

    class NotInt(object):
        pass

    class Float(object):
        def __float__(self):
            return 41.99999

    class Index(object):
        def __index__(self):
            return 42

    class IntAndIndex(object):
        def __int__(self):
            return 42

        def __index__(self):
            return 0

    class RaisingTypeErrorOnIndex(object):
        def __index__(self):
            raise TypeError

        def __int__(self):
            return 42

    class RaisingValueErrorOnIndex(object):
        def __index__(self):
            raise ValueError

        def __int__(self):
            return 42

    convert, noconvert = m.int_passthrough, m.int_passthrough_noconvert

    def requires_conversion(v):
        pytest.raises(TypeError, noconvert, v)

    def cant_convert(v):
        pytest.raises(TypeError, convert, v)

    assert convert(7) == 7
    assert noconvert(7) == 7
    cant_convert(3.14159)
    # TODO: Avoid DeprecationWarning in `PyLong_AsLong` (and similar)
    # TODO: PyPy 3.8 does not behave like CPython 3.8 here yet (7.3.7)
    if (3, 8) <= env.PY < (3, 10) and env.CPYTHON:
        with env.deprecated_call():
            assert convert(Int()) == 42
    else:
        assert convert(Int()) == 42
    requires_conversion(Int())
    cant_convert(NotInt())
    cant_convert(Float())

    # Before Python 3.8, `PyLong_AsLong` does not pick up on `obj.__index__`,
    # but pybind11 "backports" this behavior.
    assert convert(Index()) == 42
    assert noconvert(Index()) == 42
    assert convert(IntAndIndex()) == 0  # Fishy; `int(DoubleThought)` == 42
    assert noconvert(IntAndIndex()) == 0
    assert convert(RaisingTypeErrorOnIndex()) == 42
    requires_conversion(RaisingTypeErrorOnIndex())
    assert convert(RaisingValueErrorOnIndex()) == 42
    requires_conversion(RaisingValueErrorOnIndex())


def test_numpy_int_convert():
    np = pytest.importorskip("numpy")

    convert, noconvert = m.int_passthrough, m.int_passthrough_noconvert

    def require_implicit(v):
        pytest.raises(TypeError, noconvert, v)

    # `np.intc` is an alias that corresponds to a C++ `int`
    assert convert(np.intc(42)) == 42
    assert noconvert(np.intc(42)) == 42

    # The implicit conversion from np.float32 is undesirable but currently accepted.
    # TODO: Avoid DeprecationWarning in `PyLong_AsLong` (and similar)
    # TODO: PyPy 3.8 does not behave like CPython 3.8 here yet (7.3.7)
    # https://github.com/pybind/pybind11/issues/3408
    if (3, 8) <= env.PY < (3, 10) and env.CPYTHON:
        with env.deprecated_call():
            assert convert(np.float32(3.14159)) == 3
    else:
        assert convert(np.float32(3.14159)) == 3
    require_implicit(np.float32(3.14159))


def test_tuple(doc):
    """std::pair <-> tuple & std::tuple <-> tuple"""
    assert m.pair_passthrough((True, "test")) == ("test", True)
    assert m.tuple_passthrough((True, "test", 5)) == (5, "test", True)
    # Any sequence can be cast to a std::pair or std::tuple
    assert m.pair_passthrough([True, "test"]) == ("test", True)
    assert m.tuple_passthrough([True, "test", 5]) == (5, "test", True)
    assert m.empty_tuple() == ()

    assert (
        doc(m.pair_passthrough)
        == """
        pair_passthrough(arg0: Tuple[bool, str]) -> Tuple[str, bool]

        Return a pair in reversed order
    """
    )
    assert (
        doc(m.tuple_passthrough)
        == """
        tuple_passthrough(arg0: Tuple[bool, str, int]) -> Tuple[int, str, bool]

        Return a triple in reversed order
    """
    )

    assert m.rvalue_pair() == ("rvalue", "rvalue")
    assert m.lvalue_pair() == ("lvalue", "lvalue")
    assert m.rvalue_tuple() == ("rvalue", "rvalue", "rvalue")
    assert m.lvalue_tuple() == ("lvalue", "lvalue", "lvalue")
    assert m.rvalue_nested() == ("rvalue", ("rvalue", ("rvalue", "rvalue")))
    assert m.lvalue_nested() == ("lvalue", ("lvalue", ("lvalue", "lvalue")))

    assert m.int_string_pair() == (2, "items")


def test_builtins_cast_return_none():
    """Casters produced with PYBIND11_TYPE_CASTER() should convert nullptr to None"""
    assert m.return_none_string() is None
    assert m.return_none_char() is None
    assert m.return_none_bool() is None
    assert m.return_none_int() is None
    assert m.return_none_float() is None
    assert m.return_none_pair() is None


def test_none_deferred():
    """None passed as various argument types should defer to other overloads"""
    assert not m.defer_none_cstring("abc")
    assert m.defer_none_cstring(None)
    assert not m.defer_none_custom(UserType())
    assert m.defer_none_custom(None)
    assert m.nodefer_none_void(None)


def test_void_caster():
    assert m.load_nullptr_t(None) is None
    assert m.cast_nullptr_t() is None


def test_reference_wrapper():
    """std::reference_wrapper for builtin and user types"""
    assert m.refwrap_builtin(42) == 420
    assert m.refwrap_usertype(UserType(42)) == 42
    assert m.refwrap_usertype_const(UserType(42)) == 42

    with pytest.raises(TypeError) as excinfo:
        m.refwrap_builtin(None)
    assert "incompatible function arguments" in str(excinfo.value)

    with pytest.raises(TypeError) as excinfo:
        m.refwrap_usertype(None)
    assert "incompatible function arguments" in str(excinfo.value)

    assert m.refwrap_lvalue().value == 1
    assert m.refwrap_lvalue_const().value == 1

    a1 = m.refwrap_list(copy=True)
    a2 = m.refwrap_list(copy=True)
    assert [x.value for x in a1] == [2, 3]
    assert [x.value for x in a2] == [2, 3]
    assert not a1[0] is a2[0] and not a1[1] is a2[1]

    b1 = m.refwrap_list(copy=False)
    b2 = m.refwrap_list(copy=False)
    assert [x.value for x in b1] == [1, 2]
    assert [x.value for x in b2] == [1, 2]
    assert b1[0] is b2[0] and b1[1] is b2[1]

    assert m.refwrap_iiw(IncType(5)) == 5
    assert m.refwrap_call_iiw(IncType(10), m.refwrap_iiw) == [10, 10, 10, 10]


def test_complex_cast():
    """std::complex casts"""
    assert m.complex_cast(1) == "1.0"
    assert m.complex_cast(2j) == "(0.0, 2.0)"


def test_bool_caster():
    """Test bool caster implicit conversions."""
    convert, noconvert = m.bool_passthrough, m.bool_passthrough_noconvert

    def require_implicit(v):
        pytest.raises(TypeError, noconvert, v)

    def cant_convert(v):
        pytest.raises(TypeError, convert, v)

    # straight up bool
    assert convert(True) is True
    assert convert(False) is False
    assert noconvert(True) is True
    assert noconvert(False) is False

    # None requires implicit conversion
    require_implicit(None)
    assert convert(None) is False

    class A(object):
        def __init__(self, x):
            self.x = x

        def __nonzero__(self):
            return self.x

        def __bool__(self):
            return self.x

    class B(object):
        pass

    # Arbitrary objects are not accepted
    cant_convert(object())
    cant_convert(B())

    # Objects with __nonzero__ / __bool__ defined can be converted
    require_implicit(A(True))
    assert convert(A(True)) is True
    assert convert(A(False)) is False


def test_numpy_bool():
    np = pytest.importorskip("numpy")

    convert, noconvert = m.bool_passthrough, m.bool_passthrough_noconvert

    def cant_convert(v):
        pytest.raises(TypeError, convert, v)

    # np.bool_ is not considered implicit
    assert convert(np.bool_(True)) is True
    assert convert(np.bool_(False)) is False
    assert noconvert(np.bool_(True)) is True
    assert noconvert(np.bool_(False)) is False
    cant_convert(np.zeros(2, dtype="int"))


def test_int_long():
    """In Python 2, a C++ int should return a Python int rather than long
    if possible: longs are not always accepted where ints are used (such
    as the argument to sys.exit()). A C++ long long is always a Python
    long."""

    import sys

    must_be_long = type(getattr(sys, "maxint", 1) + 1)
    assert isinstance(m.int_cast(), int)
    assert isinstance(m.long_cast(), int)
    assert isinstance(m.longlong_cast(), must_be_long)


def test_void_caster_2():
    assert m.test_void_caster()


def test_const_ref_caster():
    """Verifies that const-ref is propagated through type_caster cast_op.
    The returned ConstRefCasted type is a minimal type that is constructed to
    reference the casting mode used.
    """
    x = False
    assert m.takes(x) == 1
    assert m.takes_move(x) == 1

    assert m.takes_ptr(x) == 3
    assert m.takes_ref(x) == 2
    assert m.takes_ref_wrap(x) == 2

    assert m.takes_const_ptr(x) == 5
    assert m.takes_const_ref(x) == 4
    assert m.takes_const_ref_wrap(x) == 4



# -*- coding: utf-8 -*-
import os
import subprocess
import sys
from textwrap import dedent

import pytest

DIR = os.path.abspath(os.path.dirname(__file__))
MAIN_DIR = os.path.dirname(os.path.dirname(DIR))
WIN = sys.platform.startswith("win32") or sys.platform.startswith("cygwin")


@pytest.mark.parametrize("parallel", [False, True])
@pytest.mark.parametrize("std", [11, 0])
def test_simple_setup_py(monkeypatch, tmpdir, parallel, std):
    monkeypatch.chdir(tmpdir)
    monkeypatch.syspath_prepend(MAIN_DIR)

    (tmpdir / "setup.py").write_text(
        dedent(
            u"""\
            import sys
            sys.path.append({MAIN_DIR!r})

            from setuptools import setup, Extension
            from pybind11.setup_helpers import build_ext, Pybind11Extension

            std = {std}

            ext_modules = [
                Pybind11Extension(
                    "simple_setup",
                    sorted(["main.cpp"]),
                    cxx_std=std,
                ),
            ]

            cmdclass = dict()
            if std == 0:
                cmdclass["build_ext"] = build_ext


            parallel = {parallel}
            if parallel:
                from pybind11.setup_helpers import ParallelCompile
                ParallelCompile().install()

            setup(
                name="simple_setup_package",
                cmdclass=cmdclass,
                ext_modules=ext_modules,
            )
            """
        ).format(MAIN_DIR=MAIN_DIR, std=std, parallel=parallel),
        encoding="ascii",
    )

    (tmpdir / "main.cpp").write_text(
        dedent(
            u"""\
            #include <pybind11/pybind11.h>

            int f(int x) {
                return x * 3;
            }
            PYBIND11_MODULE(simple_setup, m) {
                m.def("f", &f);
            }
            """
        ),
        encoding="ascii",
    )

    out = subprocess.check_output(
        [sys.executable, "setup.py", "build_ext", "--inplace"],
    )
    if not WIN:
        assert b"-g0" in out
    out = subprocess.check_output(
        [sys.executable, "setup.py", "build_ext", "--inplace", "--force"],
        env=dict(os.environ, CFLAGS="-g"),
    )
    if not WIN:
        assert b"-g0" not in out

    # Debug helper printout, normally hidden
    print(out)
    for item in tmpdir.listdir():
        print(item.basename)

    assert (
        len([f for f in tmpdir.listdir() if f.basename.startswith("simple_setup")]) == 1
    )
    assert len(list(tmpdir.listdir())) == 4  # two files + output + build_dir

    (tmpdir / "test.py").write_text(
        dedent(
            u"""\
            import simple_setup
            assert simple_setup.f(3) == 9
            """
        ),
        encoding="ascii",
    )

    subprocess.check_call(
        [sys.executable, "test.py"], stdout=sys.stdout, stderr=sys.stderr
    )


def test_intree_extensions(monkeypatch, tmpdir):
    monkeypatch.syspath_prepend(MAIN_DIR)

    from pybind11.setup_helpers import intree_extensions

    monkeypatch.chdir(tmpdir)
    root = tmpdir
    root.ensure_dir()
    subdir = root / "dir"
    subdir.ensure_dir()
    src = subdir / "ext.cpp"
    src.ensure()
    (ext,) = intree_extensions([src.relto(tmpdir)])
    assert ext.name == "ext"
    subdir.ensure("__init__.py")
    (ext,) = intree_extensions([src.relto(tmpdir)])
    assert ext.name == "dir.ext"


def test_intree_extensions_package_dir(monkeypatch, tmpdir):
    monkeypatch.syspath_prepend(MAIN_DIR)

    from pybind11.setup_helpers import intree_extensions

    monkeypatch.chdir(tmpdir)
    root = tmpdir / "src"
    root.ensure_dir()
    subdir = root / "dir"
    subdir.ensure_dir()
    src = subdir / "ext.cpp"
    src.ensure()
    (ext,) = intree_extensions([src.relto(tmpdir)], package_dir={"": "src"})
    assert ext.name == "dir.ext"
    (ext,) = intree_extensions([src.relto(tmpdir)], package_dir={"foo": "src"})
    assert ext.name == "foo.dir.ext"
    subdir.ensure("__init__.py")
    (ext,) = intree_extensions([src.relto(tmpdir)], package_dir={"": "src"})
    assert ext.name == "dir.ext"
    (ext,) = intree_extensions([src.relto(tmpdir)], package_dir={"foo": "src"})
    assert ext.name == "foo.dir.ext"



# IMPORTANT: Should stay in sync with setup_helpers.py (mostly checked by CI /
# pre-commit).

import contextlib
import distutils.ccompiler
from distutils.command.build_ext import build_ext as _build_ext  # type: ignore
from distutils.extension import Extension as _Extension
from types import TracebackType
from typing import Any, Callable, Dict, Iterator, List, Optional, Type, TypeVar, Union

WIN: bool
PY2: bool
MACOS: bool
STD_TMPL: str

class Pybind11Extension(_Extension):
    def _add_cflags(self, *flags: str) -> None: ...
    def _add_lflags(self, *flags: str) -> None: ...
    def __init__(
        self, *args: Any, cxx_std: int = 0, language: str = "c++", **kwargs: Any
    ) -> None: ...
    @property
    def cxx_std(self) -> int: ...
    @cxx_std.setter
    def cxx_std(self, level: int) -> None: ...

@contextlib.contextmanager
def tmp_chdir() -> Iterator[str]: ...
def has_flag(compiler: distutils.ccompiler.CCompiler, flag: str) -> bool: ...
def auto_cpp_level(compiler: distutils.ccompiler.CCompiler) -> Union[int, str]: ...

class build_ext(_build_ext):  # type: ignore
    def build_extensions(self) -> None: ...

def intree_extensions(
    paths: Iterator[str], package_dir: Optional[Dict[str, str]] = None
) -> List[Pybind11Extension]: ...
def no_recompile(obj: str, src: str) -> bool: ...
def naive_recompile(obj: str, src: str) -> bool: ...

T = TypeVar("T", bound="ParallelCompile")

class ParallelCompile:
    envvar: Optional[str]
    default: int
    max: int
    needs_recompile: Callable[[str, str], bool]
    def __init__(
        self,
        envvar: Optional[str] = None,
        default: int = 0,
        max: int = 0,
        needs_recompile: Callable[[str, str], bool] = no_recompile,
    ) -> None: ...
    def function(self) -> Any: ...
    def install(self: T) -> T: ...
    def __enter__(self: T) -> T: ...
    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None: ...



# -*- coding: utf-8 -*-
import pytest

import env  # noqa: F401
from pybind11_tests import local_bindings as m


def test_load_external():
    """Load a `py::module_local` type that's only registered in an external module"""
    import pybind11_cross_module_tests as cm

    assert m.load_external1(cm.ExternalType1(11)) == 11
    assert m.load_external2(cm.ExternalType2(22)) == 22

    with pytest.raises(TypeError) as excinfo:
        assert m.load_external2(cm.ExternalType1(21)) == 21
    assert "incompatible function arguments" in str(excinfo.value)

    with pytest.raises(TypeError) as excinfo:
        assert m.load_external1(cm.ExternalType2(12)) == 12
    assert "incompatible function arguments" in str(excinfo.value)


def test_local_bindings():
    """Tests that duplicate `py::module_local` class bindings work across modules"""

    # Make sure we can load the second module with the conflicting (but local) definition:
    import pybind11_cross_module_tests as cm

    i1 = m.LocalType(5)
    assert i1.get() == 4
    assert i1.get3() == 8

    i2 = cm.LocalType(10)
    assert i2.get() == 11
    assert i2.get2() == 12

    assert not hasattr(i1, "get2")
    assert not hasattr(i2, "get3")

    # Loading within the local module
    assert m.local_value(i1) == 5
    assert cm.local_value(i2) == 10

    # Cross-module loading works as well (on failure, the type loader looks for
    # external module-local converters):
    assert m.local_value(i2) == 10
    assert cm.local_value(i1) == 5


def test_nonlocal_failure():
    """Tests that attempting to register a non-local type in multiple modules fails"""
    import pybind11_cross_module_tests as cm

    with pytest.raises(RuntimeError) as excinfo:
        cm.register_nonlocal()
    assert (
        str(excinfo.value) == 'generic_type: type "NonLocalType" is already registered!'
    )


def test_duplicate_local():
    """Tests expected failure when registering a class twice with py::local in the same module"""
    with pytest.raises(RuntimeError) as excinfo:
        m.register_local_external()
    import pybind11_tests

    assert str(excinfo.value) == (
        'generic_type: type "LocalExternal" is already registered!'
        if hasattr(pybind11_tests, "class_")
        else "test_class not enabled"
    )


def test_stl_bind_local():
    import pybind11_cross_module_tests as cm

    v1, v2 = m.LocalVec(), cm.LocalVec()
    v1.append(m.LocalType(1))
    v1.append(m.LocalType(2))
    v2.append(cm.LocalType(1))
    v2.append(cm.LocalType(2))

    # Cross module value loading:
    v1.append(cm.LocalType(3))
    v2.append(m.LocalType(3))

    assert [i.get() for i in v1] == [0, 1, 2]
    assert [i.get() for i in v2] == [2, 3, 4]

    v3, v4 = m.NonLocalVec(), cm.NonLocalVec2()
    v3.append(m.NonLocalType(1))
    v3.append(m.NonLocalType(2))
    v4.append(m.NonLocal2(3))
    v4.append(m.NonLocal2(4))

    assert [i.get() for i in v3] == [1, 2]
    assert [i.get() for i in v4] == [13, 14]

    d1, d2 = m.LocalMap(), cm.LocalMap()
    d1["a"] = v1[0]
    d1["b"] = v1[1]
    d2["c"] = v2[0]
    d2["d"] = v2[1]
    assert {i: d1[i].get() for i in d1} == {"a": 0, "b": 1}
    assert {i: d2[i].get() for i in d2} == {"c": 2, "d": 3}


def test_stl_bind_global():
    import pybind11_cross_module_tests as cm

    with pytest.raises(RuntimeError) as excinfo:
        cm.register_nonlocal_map()
    assert (
        str(excinfo.value) == 'generic_type: type "NonLocalMap" is already registered!'
    )

    with pytest.raises(RuntimeError) as excinfo:
        cm.register_nonlocal_vec()
    assert (
        str(excinfo.value) == 'generic_type: type "NonLocalVec" is already registered!'
    )

    with pytest.raises(RuntimeError) as excinfo:
        cm.register_nonlocal_map2()
    assert (
        str(excinfo.value) == 'generic_type: type "NonLocalMap2" is already registered!'
    )


def test_mixed_local_global():
    """Local types take precedence over globally registered types: a module with a `module_local`
    type can be registered even if the type is already registered globally.  With the module,
    casting will go to the local type; outside the module casting goes to the global type."""
    import pybind11_cross_module_tests as cm

    m.register_mixed_global()
    m.register_mixed_local()

    a = []
    a.append(m.MixedGlobalLocal(1))
    a.append(m.MixedLocalGlobal(2))
    a.append(m.get_mixed_gl(3))
    a.append(m.get_mixed_lg(4))

    assert [x.get() for x in a] == [101, 1002, 103, 1004]

    cm.register_mixed_global_local()
    cm.register_mixed_local_global()
    a.append(m.MixedGlobalLocal(5))
    a.append(m.MixedLocalGlobal(6))
    a.append(cm.MixedGlobalLocal(7))
    a.append(cm.MixedLocalGlobal(8))
    a.append(m.get_mixed_gl(9))
    a.append(m.get_mixed_lg(10))
    a.append(cm.get_mixed_gl(11))
    a.append(cm.get_mixed_lg(12))

    assert [x.get() for x in a] == [
        101,
        1002,
        103,
        1004,
        105,
        1006,
        207,
        2008,
        109,
        1010,
        211,
        2012,
    ]


def test_internal_locals_differ():
    """Makes sure the internal local type map differs across the two modules"""
    import pybind11_cross_module_tests as cm

    assert m.local_cpp_types_addr() != cm.local_cpp_types_addr()


@pytest.mark.xfail("env.PYPY and sys.pypy_version_info < (7, 3, 2)")
def test_stl_caster_vs_stl_bind(msg):
    """One module uses a generic vector caster from `<pybind11/stl.h>` while the other
    exports `std::vector<int>` via `py:bind_vector` and `py::module_local`"""
    import pybind11_cross_module_tests as cm

    v1 = cm.VectorInt([1, 2, 3])
    assert m.load_vector_via_caster(v1) == 6
    assert cm.load_vector_via_binding(v1) == 6

    v2 = [1, 2, 3]
    assert m.load_vector_via_caster(v2) == 6
    with pytest.raises(TypeError) as excinfo:
        cm.load_vector_via_binding(v2)
    assert (
        msg(excinfo.value)
        == """
    load_vector_via_binding(): incompatible function arguments. The following argument types are supported:
        1. (arg0: pybind11_cross_module_tests.VectorInt) -> int

    Invoked with: [1, 2, 3]
    """  # noqa: E501 line too long
    )


def test_cross_module_calls():
    import pybind11_cross_module_tests as cm

    v1 = m.LocalVec()
    v1.append(m.LocalType(1))
    v2 = cm.LocalVec()
    v2.append(cm.LocalType(2))

    # Returning the self pointer should get picked up as returning an existing
    # instance (even when that instance is of a foreign, non-local type).
    assert m.return_self(v1) is v1
    assert cm.return_self(v2) is v2
    assert m.return_self(v2) is v2
    assert cm.return_self(v1) is v1

    assert m.LocalVec is not cm.LocalVec
    # Returning a copy, on the other hand, always goes to the local type,
    # regardless of where the source type came from.
    assert type(m.return_copy(v1)) is m.LocalVec
    assert type(m.return_copy(v2)) is m.LocalVec
    assert type(cm.return_copy(v1)) is cm.LocalVec
    assert type(cm.return_copy(v2)) is cm.LocalVec

    # Test the example given in the documentation (which also tests inheritance casting):
    mycat = m.Cat("Fluffy")
    mydog = cm.Dog("Rover")
    assert mycat.get_name() == "Fluffy"
    assert mydog.name() == "Rover"
    assert m.Cat.__base__.__name__ == "Pet"
    assert cm.Dog.__base__.__name__ == "Pet"
    assert m.Cat.__base__ is not cm.Dog.__base__
    assert m.pet_name(mycat) == "Fluffy"
    assert m.pet_name(mydog) == "Rover"
    assert cm.pet_name(mycat) == "Fluffy"
    assert cm.pet_name(mydog) == "Rover"

    assert m.MixGL is not cm.MixGL
    a = m.MixGL(1)
    b = cm.MixGL(2)
    assert m.get_gl_value(a) == 11
    assert m.get_gl_value(b) == 12
    assert cm.get_gl_value(a) == 101
    assert cm.get_gl_value(b) == 102

    c, d = m.MixGL2(3), cm.MixGL2(4)
    with pytest.raises(TypeError) as excinfo:
        m.get_gl_value(c)
    assert "incompatible function arguments" in str(excinfo.value)
    with pytest.raises(TypeError) as excinfo:
        m.get_gl_value(d)
    assert "incompatible function arguments" in str(excinfo.value)



# poppler-python: python binding to the poppler-cpp pdf lib
# Copyright (C) 2020, Charles Brunet <charles@cbrunet.net>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


class TocItem:
    def __init__(self, toc_item):
        self._item = toc_item

    def children(self):
        return [TocItem(i) for i in self._item]

    def __iter__(self):
        for item in self._item:
            yield TocItem(item)

    @property
    def is_open(self):
        return self._item.is_open()

    @property
    def title(self):
        return str(self._item.title())


class Toc:
    def __init__(self, toc):
        self._toc = toc

    @property
    def root(self):
        return TocItem(self._toc.root())



# poppler-python: python binding to the poppler-cpp pdf lib
# Copyright (C) 2020, Charles Brunet <charles@cbrunet.net>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from poppler._version import *  # noqa

from poppler.utilities import version
from poppler.cpp.version import version_string
from poppler.document import load, load_from_file, load_from_data
from poppler.pagerenderer import PageRenderer
from poppler.rectangle import Rectangle

from poppler.cpp.document import page_layout_enum as PageLayout
from poppler.cpp.document import page_mode_enum as PageMode

from poppler.cpp.font import type_enum as FontType

from poppler.cpp.global_ import case_sensitivity_enum as CaseSensitivity
from poppler.cpp.global_ import page_box_enum as PageBox
from poppler.cpp.global_ import permission_enum as Permission
from poppler.cpp.global_ import rotation_enum as Rotation

from poppler.cpp.image import format_enum as ImageFormat

from poppler.cpp.page import orientation_enum as PageOrientation
from poppler.cpp.page import search_direction_enum as SearchDirection
from poppler.cpp.page import text_layout_enum as TextLayout

from poppler.cpp.page_renderer import render_hint as RenderHint

from poppler.cpp.page_transition import alignment_enum as Alignment
from poppler.cpp.page_transition import direction_enum as Direction
from poppler.cpp.page_transition import type_enum as TransitionType


__all__ = [
    "version",
    "version_string",
    "load",
    "load_from_data",
    "load_from_file",
    "PageRenderer",
    "Rectangle",
    "Alignment",
    "CaseSensitivity",
    "Direction",
    "FontType",
    "ImageFormat",
    "PageBox",
    "PageLayout",
    "PageMode",
    "PageOrientation",
    "Permission",
    "RenderHint",
    "Rotation",
    "SearchDirection",
    "TextLayout",
    "TransitionType",
]

if version() >= (0, 65, 0):
    from poppler.cpp.page_renderer import line_mode_enum as LineMode  # noqa

    __all__.append("LineMode")

if version() >= (0, 73, 0):
    from poppler.cpp.global_ import set_data_dir  # noqa

    __all__.append("set_data_dir")

if version() >= (0, 74, 0):
    from poppler.cpp.destination import type_enum as DestinationType  # noqa

    __all__.append("DestinationType")



# poppler-python: python binding to the poppler-cpp pdf lib
# Copyright (C) 2020, Charles Brunet <charles@cbrunet.net>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from poppler.cpp import image
from poppler.rectangle import Rectangle


class Image:

    Format = image.format_enum

    def __init__(
        self, width=0, height=0, iformat=Format.invalid, data=None, poppler_object=None
    ):
        if poppler_object is not None:
            self._image = poppler_object
        elif data is not None:
            self._image = image.image(data, width, height, iformat)
        elif width > 0:
            self._image = image.image(width, height, iformat)
        else:
            self._image = image.image()

    @classmethod
    def from_object(cls, poppler_object):
        return cls(poppler_object=poppler_object)

    @property
    def bytes_per_row(self):
        return self._image.bytes_per_row()

    @property
    def const_data(self):
        return self._image.data()

    def copy(self, rect=None):
        image = self._image.copy(rect or Rectangle()._rect)
        return Image.from_object(image)

    @property
    def data(self):
        return self._image.data()

    @data.setter
    def data(self, data):
        self._image.set_data(data)

    @property
    def format(self):
        return self._image.format()

    @property
    def height(self):
        return self._image.height()

    @property
    def is_valid(self):
        return self._image.is_valid()

    def save(self, file_name, out_format, dpi=-1):
        return self._image.save(str(file_name), out_format, dpi)

    @property
    def width(self):
        return self._image.width()

    def memoryview(self):
        return memoryview(self._image)

    @staticmethod
    def supported_image_formats():
        return image.supported_image_formats()



# poppler-python: python binding to the poppler-cpp pdf lib
# Copyright (C) 2020, Charles Brunet <charles@cbrunet.net>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from poppler.utilities import from_time_type


class EmbeddedFile:
    def __init__(self, file):
        self._file = file

    @property
    def checksum(self):
        return self._file.checksum()

    @property
    def creation_date(self):
        return from_time_type(self._file.creation_date())

    @property
    def data(self):
        return self._file.data()

    @property
    def description(self):
        return str(self._file.description())

    @property
    def is_valid(self):
        return self._file.is_valid()

    @property
    def mime_type(self):
        return self._file.mime_type()

    @property
    def modification_date(self):
        return from_time_type(self._file.modification_date())

    @property
    def name(self):
        return self._file.name()

    @property
    def size(self):
        return self._file.size()



# poppler-python: python binding to the poppler-cpp pdf lib
# Copyright (C) 2020, Charles Brunet <charles@cbrunet.net>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from poppler.utilities import version, since

if version() >= (0, 74, 0):
    from poppler.cpp.destination import type_enum as DestinationType  # noqa


@since(0, 74)
class Destination:

    def __init__(self, destination):
        self._destination = destination

    @property
    def bottom(self):
        return self._destination.bottom()

    @property
    def is_change_left(self):
        return self._destination.is_change_left()

    @property
    def is_change_top(self):
        return self._destination.is_change_top()

    @property
    def is_change_zoom(self):
        return self._destination.is_change_zoom()

    @property
    def left(self):
        return self._destination.left()

    @property
    def page_number(self):
        return self._destination.page_number()

    @property
    def right(self):
        return self._destination.right()

    @property
    def top(self):
        return self._destination.top()

    @property
    def type(self):
        return self._destination.type()

    @property
    def zoom(self):
        return self._destination.zoom()



# poppler-python: python binding to the poppler-cpp pdf lib
# Copyright (C) 2020, Charles Brunet <charles@cbrunet.net>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from poppler.cpp.font import font_info
from poppler.cpp.font import type_enum as FontType  # noqa


class FontInfo:
    def __init__(self, info=None):
        self._info = info or font_info()

    @property
    def file(self):
        return self._info.file()

    @property
    def is_embedded(self):
        return self._info.is_embedded()

    @property
    def is_subset(self):
        return self._info.is_subset()

    @property
    def name(self):
        return self._info.name()

    @property
    def type(self):
        return self._info.type()


class FontIterator:
    def __init__(self, it):
        self._it = it

    @property
    def current_page(self):
        return self._it.current_page()

    @property
    def has_next(self):
        return self._it.has_next()

    def next_(self):
        return [FontInfo(i) for i in self._it.next_()]

    def __iter__(self):
        while self._it.has_next():
            yield self.current_page, self.next_()



# poppler-python: python binding to the poppler-cpp pdf lib
# Copyright (C) 2020, Charles Brunet <charles@cbrunet.net>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import pytest

from poppler import version

if version() >= (0, 74, 0):
    from poppler import DestinationType

pytestmark = pytest.mark.skipif(
    version() < (0, 74, 0), reason="Requires at least Poppler 0.74.0"
)


def test_destination_map_keys(sample_document):
    destinations = sample_document.create_destination_map()

    assert "Doc-Start" in destinations
    assert "Navigation1" in destinations
    assert "Outline0.1" in destinations
    assert "page.1" in destinations


def test_destination_map_page(sample_document):
    destinations = sample_document.create_destination_map()

    for i in range(1, 10):
        k = f"page.{i}"
        assert destinations[k].page_number == i


def test_destination_map_properties(sample_document):
    destinations = sample_document.create_destination_map()
    destination = destinations["Outline0.1"]

    assert destination.bottom == 0.0
    assert destination.top == 245.3
    assert destination.left == 28.346
    assert destination.right == 0.0
    assert destination.zoom == 0.0
    assert destination.type == DestinationType.xyz
    assert destination.is_change_left
    assert destination.is_change_top
    assert not destination.is_change_zoom
    assert destination.page_number == 2



# poppler-python: python binding to the poppler-cpp pdf lib
# Copyright (C) 2020, Charles Brunet <charles@cbrunet.net>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import pytest

from poppler.pagerenderer import PageRenderer, RenderHint
from poppler.image import Image
from poppler import version

if version() >= (0, 65, 0):
    from poppler.pagerenderer import LineMode


def test_can_render():
    assert PageRenderer.can_render() is True


@pytest.mark.skipif(version() < (0, 65, 0), reason="Requires at least Poppler 0.65.0")
def test_image_format():
    renderer = PageRenderer()

    assert renderer.image_format == Image.Format.argb32


@pytest.mark.skipif(version() < (0, 65, 0), reason="Requires at least Poppler 0.65.0")
def test_set_image_format():
    renderer = PageRenderer()
    renderer.image_format = Image.Format.mono

    assert renderer.image_format == Image.Format.mono


@pytest.mark.skipif(version() < (0, 65, 0), reason="Requires at least Poppler 0.65.0")
def test_line_mode():
    renderer = PageRenderer()

    assert renderer.line_mode == LineMode.default


@pytest.mark.skipif(version() < (0, 65, 0), reason="Requires at least Poppler 0.65.0")
def test_set_line_mode():
    renderer = PageRenderer()
    renderer.line_mode = LineMode.solid

    assert renderer.line_mode == LineMode.solid


def test_paper_color():
    renderer = PageRenderer()

    assert renderer.paper_color == 0xFFFFFFFF


def test_set_paper_color():
    renderer = PageRenderer()
    renderer.paper_color = 0x12345678

    assert renderer.paper_color == 0x12345678


def test_render_hints():
    renderer = PageRenderer()

    assert renderer.render_hints == 0


def test_set_render_hints():
    renderer = PageRenderer()
    renderer.render_hints = (
        RenderHint.antialiasing | RenderHint.text_antialiasing | RenderHint.text_hinting
    )

    assert renderer.render_hints == 7


def test_set_render_hint():
    renderer = PageRenderer()
    renderer.render_hints = 7
    renderer.set_render_hint(RenderHint.antialiasing, False)

    assert renderer.render_hints == 6


def test_render_page(pdf_page):
    renderer = PageRenderer()
    image = renderer.render_page(pdf_page)

    assert image.height == 792
    assert image.width == 612


def test_image_with_paper_color(pdf_page):
    renderer = PageRenderer()
    renderer.paper_color = 0xFF336699
    image = renderer.render_page(pdf_page)

    assert image.data[0:4] == b"\x99\x66\x33\xFF"



# -*- coding: utf-8 -*-

import gc
import weakref

import pytest

import env  # noqa: F401
from pybind11_tests import custom_type_setup as m


@pytest.fixture
def gc_tester():
    """Tests that an object is garbage collected.

    Assumes that any unreferenced objects are fully collected after calling
    `gc.collect()`.  That is true on CPython, but does not appear to reliably
    hold on PyPy.
    """

    weak_refs = []

    def add_ref(obj):
        # PyPy does not support `gc.is_tracked`.
        if hasattr(gc, "is_tracked"):
            assert gc.is_tracked(obj)
        weak_refs.append(weakref.ref(obj))

    yield add_ref

    gc.collect()
    for ref in weak_refs:
        assert ref() is None


# PyPy does not seem to reliably garbage collect.
@pytest.mark.skipif("env.PYPY")
def test_self_cycle(gc_tester):
    obj = m.OwnsPythonObjects()
    obj.value = obj
    gc_tester(obj)


# PyPy does not seem to reliably garbage collect.
@pytest.mark.skipif("env.PYPY")
def test_indirect_cycle(gc_tester):
    obj = m.OwnsPythonObjects()
    obj_list = [obj]
    obj.value = obj_list
    gc_tester(obj)



# poppler-python: python binding to the poppler-cpp pdf lib
# Copyright (C) 2020, Charles Brunet <charles@cbrunet.net>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


from poppler.cpp import page_transition


class PageTransition:

    Alignment = page_transition.alignment_enum
    Direction = page_transition.direction_enum
    Type = page_transition.type_enum

    def __init__(self, transition):
        self._transition = transition

    @property
    def alignment(self):
        return self._transition.alignment()

    @property
    def angle(self):
        return self._transition.angle()

    @property
    def direction(self):
        return self._transition.direction()

    @property
    def duration(self):
        return self._transition.duration()

    @property
    def is_rectangular(self):
        return self._transition.is_rectangular()

    @property
    def scale(self):
        return self._transition.scale()

    @property
    def type(self):
        return self._transition.type()



# poppler-python: python binding to the poppler-cpp pdf lib
# Copyright (C) 2020, Charles Brunet <charles@cbrunet.net>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from poppler.font import FontInfo, FontType


def test_invalid_font_info():
    font_info = FontInfo()

    assert font_info.type == FontType.unknown


def test_font_iterator(sample_document):
    font_iterator = sample_document.create_font_iterator()

    assert font_iterator.has_next
    assert font_iterator.current_page == 0

    fonts = font_iterator.next_()

    assert len(fonts) == 3
    assert font_iterator.current_page == 1


def test_font_iterator_interface(sample_document):
    font_iterator = sample_document.create_font_iterator()

    for i, (page, _) in enumerate(font_iterator):
        assert i == page


def test_font_info(sample_document):
    fonts = sample_document.fonts()
    font = fonts[0]

    assert font.file == ""
    assert font.is_embedded
    assert font.is_subset
    assert font.name == "SBNSFQ+CMSS8"
    assert font.type == FontType.type1



import os
import re
import sys
from pathlib import Path
import platform
import subprocess

from setuptools import setup, Extension, find_packages
from setuptools.command.build_ext import build_ext
from distutils.version import LooseVersion

exec(Path("src/poppler/_version.py").read_text())


class CMakeExtension(Extension):
    def __init__(self, name, sourcedir=""):
        Extension.__init__(self, name, sources=[])
        self.sourcedir = os.path.abspath(sourcedir)


class CMakeBuild(build_ext):
    def run(self):
        try:
            out = subprocess.check_output(["cmake", "--version"])
        except OSError:
            raise RuntimeError(
                "CMake must be installed to build the following extensions: "
                + ", ".join(e.name for e in self.extensions)
            )

        if platform.system() == "Windows":
            cmake_version = LooseVersion(
                re.search(r"version\s*([\d.]+)", out.decode()).group(1)
            )
            if cmake_version < "3.1.0":
                raise RuntimeError("CMake >= 3.1.0 is required on Windows")

        for ext in self.extensions:
            self.build_extension(ext)

    def build_extension(self, ext):
        extdir = os.path.abspath(os.path.dirname(self.get_ext_fullpath(ext.name)))
        cmake_args = [
            "-DCMAKE_LIBRARY_OUTPUT_DIRECTORY=" + extdir,
            "-DPYTHON_EXECUTABLE=" + sys.executable,
        ]

        cfg = "Debug" if self.debug else "Release"
        build_args = ["--config", cfg]

        if platform.system() == "Windows":
            cmake_args += [
                "-DCMAKE_LIBRARY_OUTPUT_DIRECTORY_{}={}".format(cfg.upper(), extdir)
            ]
            if sys.maxsize > 2 ** 32:
                cmake_args += ["-A", "x64"]
            build_args += ["--", "/m"]
        else:
            cmake_args += ["-DCMAKE_BUILD_TYPE=" + cfg]
            build_args += ["--", "-j2"]

        env = os.environ.copy()
        env["CXXFLAGS"] = '{} -DVERSION_INFO=\\"{}\\"'.format(
            env.get("CXXFLAGS", ""), self.distribution.get_version()
        )
        if not os.path.exists(self.build_temp):
            os.makedirs(self.build_temp)
        subprocess.check_call(
            ["cmake", ext.sourcedir] + cmake_args, cwd=self.build_temp, env=env
        )
        subprocess.check_call(
            ["cmake", "--build", "."] + build_args, cwd=self.build_temp
        )


setup(
    name="python-poppler",
    version=__version__,  # noqa
    author=__author__,  # noqa
    author_email="charles@cbrunet.net",
    url="https://github.com/cbrunet/python-poppler",
    description="A Python binding to poppler-cpp",
    long_description=(Path(__file__).parent / "README.md").read_text(),
    long_description_content_type="text/markdown",
    license="GPLv2",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    keywords=["pdf", "poppler"],
    project_urls={
        "Documentation": "https://cbrunet.github.io/python-poppler",
        "Source": "https://github.com/cbrunet/python-poppler",
        "Tracker": "https://github.com/cbrunet/python-poppler/issues",
    },
    python_requires=">=3.6",
    packages=find_packages("src"),
    package_dir={"": "src"},
    include_package_data=True,
    ext_modules=[CMakeExtension("poppler.cpp.modules")],
    cmdclass=dict(build_ext=CMakeBuild),
    zip_safe=False,
)



# poppler-python: python binding to the poppler-cpp pdf lib
# Copyright (C) 2020, Charles Brunet <charles@cbrunet.net>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import pytest

from poppler import version, version_string
from poppler.utilities import since


def test_version_string():
    assert tuple(map(int, version_string().split("."))) == version()


def test_since_fct():
    @since(99, 99)
    def fct():
        pass

    with pytest.raises(NotImplementedError):
        fct()


def test_since_cls():
    @since(99, 99)
    class cls:
        pass

    with pytest.raises(NotImplementedError):
        cls()



# poppler-python: python binding to the poppler-cpp pdf lib
# Copyright (C) 2020, Charles Brunet <charles@cbrunet.net>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from poppler import Permission
from poppler import document
from poppler import version
from poppler.document import LockedDocumentError

import pytest

from datetime import datetime, timezone


@pytest.fixture()
def locked_document(data_path):
    return document.load_from_file(data_path / "document.pdf")


def test_load_from_data(data_path):
    file_data = (data_path / "document.pdf").read_bytes()
    pdf_document = document.load_from_data(file_data, "owner", "user")
    if version() < (0, 46, 0):
        assert pdf_document.author == "Charles"
    else:
        assert pdf_document.author == "Charles Brunet"


def test_load_with_path(data_path):
    pdf_document = document.load(data_path / "document.pdf", "owner", "user")
    if version() < (0, 46, 0):
        assert pdf_document.author == "Charles"
    else:
        assert pdf_document.author == "Charles Brunet"


def test_load_with_filename(data_path):
    pdf_document = document.load(str(data_path / "document.pdf"), "owner", "user")
    if version() < (0, 46, 0):
        assert pdf_document.author == "Charles"
    else:
        assert pdf_document.author == "Charles Brunet"


def test_load_with_bytes(data_path):
    data = (data_path / "document.pdf").read_bytes()
    pdf_document = document.load(data, "owner", "user")
    if version() < (0, 46, 0):
        assert pdf_document.author == "Charles"
    else:
        assert pdf_document.author == "Charles Brunet"


def test_load_with_file(data_path):
    with (data_path / "document.pdf").open("rb") as f:
        pdf_document = document.load(f, "owner", "user")
    if version() < (0, 46, 0):
        assert pdf_document.author == "Charles"
    else:
        assert pdf_document.author == "Charles Brunet"


def test_load_with_file_not_bytes(data_path):
    with (data_path / "document.pdf").open("r") as f:
        with pytest.raises(TypeError):
            document.load(f, "owner", "user")


def test_load_with_invalid_type():
    with pytest.raises(TypeError):
        document.load(42)


def test_load_not_a_pdf_document(data_path):
    with pytest.raises(ValueError):
        _ = document.load(str(data_path / "sample.tex"))


@pytest.mark.skipif(version() < (0, 46, 0), reason="Requires at least Poppler 0.46.0")
def test_save(pdf_document, tmp_path):
    copy_document = tmp_path / "copy.pdf"
    pdf_document.author = "Valérie Tremblay"
    assert pdf_document.save(copy_document)

    pdf_copy = document.load(copy_document, "owner", "user")
    assert pdf_copy.author == "Valérie Tremblay"


@pytest.mark.skipif(version() < (0, 46, 0), reason="Requires at least Poppler 0.46.0")
def test_save_a_copy(pdf_document, tmp_path):
    copy_document = tmp_path / "copy.pdf"
    pdf_document.author = "Valérie Tremblay"
    assert pdf_document.save_a_copy(copy_document)

    pdf_copy = document.load(copy_document, "owner", "user")
    assert pdf_copy.author == "Charles Brunet"


def test_pages(pdf_document):
    assert pdf_document.pages == 3


def test_embedded_file(pdf_document):
    assert pdf_document.embedded_files() == []


def test_get_author(pdf_document):
    if version() < (0, 46, 0):
        assert pdf_document.author == "Charles"
    else:
        assert pdf_document.author == "Charles Brunet"


def test_get_creation_date(pdf_document):
    date = pdf_document.creation_date

    if version() < (0, 46, 0):
        assert date.astimezone(timezone.utc) == datetime(
            2020, 3, 25, 21, 19, 50, tzinfo=timezone.utc
        )
    else:
        assert date.astimezone(timezone.utc) == datetime(
            2020, 3, 26, 1, 19, 50, tzinfo=timezone.utc
        )


def test_get_creator(pdf_document):
    if version() < (0, 46, 0):
        assert pdf_document.creator == "Wri"
    else:
        assert pdf_document.creator == "Writer"


def test_get_keywords(pdf_document):
    assert pdf_document.keywords == ""


def test_get_modification_date(pdf_document):
    assert pdf_document.modification_date is None


def test_get_pdf_id(pdf_document):
    pdf_id = pdf_document.pdf_id
    assert pdf_id.permanent_id == "c2a52b07e1fd69cc3d4fed8a2ea67ab6"
    assert pdf_id.update_id == "c2a52b07e1fd69cc3d4fed8a2ea67ab6"


def test_get_pdf_version(pdf_document):
    version = pdf_document.pdf_version
    assert version == (1, 5)


def test_get_producer(pdf_document):
    if version() < (0, 46, 0):
        assert pdf_document.producer == "LibreOf"
    else:
        assert pdf_document.producer == "LibreOffice 6.4"


def test_get_subject(pdf_document):
    assert pdf_document.subject == ""


def test_get_title(pdf_document):
    assert pdf_document.title == ""


def test_has_embedded_files(pdf_document):
    assert pdf_document.has_embedded_files() is False
    assert not pdf_document.embedded_files()


def test_has_permission(pdf_document):
    assert pdf_document.has_permission(Permission.print) is True
    assert pdf_document.has_permission(Permission.change) is True
    assert pdf_document.has_permission(Permission.copy) is True
    assert pdf_document.has_permission(Permission.add_notes) is True
    assert pdf_document.has_permission(Permission.fill_forms) is True
    assert pdf_document.has_permission(Permission.accessibility) is True
    assert pdf_document.has_permission(Permission.assemble) is True
    assert pdf_document.has_permission(Permission.print_high_resolution) is True


def test_info_date(pdf_document):
    date = pdf_document.info_date("CreationDate")
    if version() < (0, 46, 0):
        assert date.astimezone(timezone.utc) == datetime(
            2020, 3, 25, 21, 19, 50, tzinfo=timezone.utc
        )
    else:
        assert date.astimezone(timezone.utc) == datetime(
            2020, 3, 26, 1, 19, 50, tzinfo=timezone.utc
        )


def test_info_key(pdf_document):
    info = pdf_document.info_key("Author")
    if version() < (0, 46, 0):
        assert info == "Charles"
    else:
        assert info == "Charles Brunet"


def test_info_keys(pdf_document):
    keys = pdf_document.info_keys()
    assert keys == ["Author", "Creator", "Producer", "CreationDate"]


def test_is_encrypted(pdf_document):
    assert pdf_document.is_encrypted() is True


def test_is_linearized(pdf_document):
    assert pdf_document.is_linearized() is False


def test_is_locked(pdf_document):
    assert pdf_document.is_locked() is False


def test_metadata(pdf_document):
    meta = pdf_document.metadata
    assert meta == ""


@pytest.mark.skipif(version() < (0, 46, 0), reason="Requires at least Poppler 0.46.0")
def test_set_author(pdf_document):
    author = "Valérie Tremblay"
    pdf_document.author = author
    assert pdf_document.author == author


@pytest.mark.skipif(version() < (0, 46, 0), reason="Requires at least Poppler 0.46.0")
def test_set_creation_date(pdf_document):
    d = datetime(1980, 7, 19, 10, 30, 50)
    pdf_document.creation_date = d
    assert pdf_document.creation_date == d


@pytest.mark.skipif(version() < (0, 46, 0), reason="Requires at least Poppler 0.46.0")
def test_set_empty_creation_date(pdf_document):
    pdf_document.creation_date = None
    assert "CreationDate" not in pdf_document.info_keys()


@pytest.mark.skipif(version() < (0, 46, 0), reason="Requires at least Poppler 0.46.0")
def test_set_creator(pdf_document):
    creator = "Me"
    pdf_document.creator = creator
    assert pdf_document.creator == creator


@pytest.mark.skipif(version() < (0, 46, 0), reason="Requires at least Poppler 0.46.0")
def test_set_keywords(pdf_document):
    keywords = "one, two, three"
    pdf_document.keywords = keywords
    assert pdf_document.keywords == keywords


@pytest.mark.skipif(version() < (0, 46, 0), reason="Requires at least Poppler 0.46.0")
def test_set_modification_date(pdf_document):
    d = datetime(1980, 7, 19, 10, 30, 50)
    pdf_document.modification_date = d
    assert pdf_document.modification_date == d


@pytest.mark.skipif(version() < (0, 46, 0), reason="Requires at least Poppler 0.46.0")
def test_set_producer(pdf_document):
    producer = "Me"
    pdf_document.producer = producer
    assert pdf_document.producer == producer


@pytest.mark.skipif(version() < (0, 46, 0), reason="Requires at least Poppler 0.46.0")
def test_set_subject(pdf_document):
    subject = "Me"
    pdf_document.subject = subject
    assert pdf_document.subject == subject


@pytest.mark.skipif(version() < (0, 46, 0), reason="Requires at least Poppler 0.46.0")
def test_set_title(pdf_document):
    title = "The document title"
    pdf_document.title = title
    assert pdf_document.title == title


@pytest.mark.skipif(version() < (0, 46, 0), reason="Requires at least Poppler 0.46.0")
def test_set_info_date(pdf_document):
    d = datetime(1980, 7, 19, 10, 30, 50)
    assert pdf_document.set_info_date("CreationDate", d) is True
    assert pdf_document.creation_date == d


@pytest.mark.skipif(version() < (0, 46, 0), reason="Requires at least Poppler 0.46.0")
def test_set_info_key(pdf_document):
    author = "Valérie Tremblay"
    assert pdf_document.set_info_key("Author", author) is True
    assert pdf_document.author == author


def test_infos(pdf_document):
    infos = pdf_document.infos()
    if version() < (0, 46, 0):
        assert infos['Author'] == 'Charles'
    else:
        assert infos["Author"] == "Charles Brunet"


@pytest.mark.skipif(version() < (0, 46, 0), reason="Requires at least Poppler 0.46.0")
def test_remove_info(pdf_document):
    pdf_document.remove_info()
    assert not pdf_document.infos()


def test_page_layout(pdf_document):
    layout = pdf_document.page_layout
    assert layout == pdf_document.PageLayout.no_layout


def test_page_mode(pdf_document):
    mode = pdf_document.page_mode
    assert mode == pdf_document.PageMode.use_none


def test_locked_document(locked_document):
    assert locked_document.is_locked()


def test_unlock(locked_document):
    result = locked_document.unlock("owner", "user")
    assert result is False
    assert result is locked_document.is_locked()


def test_unlock_with_owner(locked_document):
    result = locked_document.unlock("owner", "")
    assert result is False
    assert result is locked_document.is_locked()


def test_unlock_with_user(locked_document):
    result = locked_document.unlock("", "user")
    assert result is False
    assert result is locked_document.is_locked()


def test_unlock_with_wrong_passwords(locked_document):
    result = locked_document.unlock("abc", "def")
    assert result is True
    assert result is locked_document.is_locked()


@pytest.mark.skipif(version() < (0, 74, 0), reason="Requires at least Poppler 0.74.0")
def test_create_destination_map(pdf_document):
    assert not pdf_document.create_destination_map()


def test_create_toc(pdf_document):
    assert pdf_document.create_toc() is None


@pytest.mark.parametrize("page", range(3))
def test_create_font_iterator(pdf_document, page):
    font_iterator = pdf_document.create_font_iterator(page)

    assert font_iterator.current_page == page


def test_fonts(pdf_document):
    fonts = pdf_document.fonts()

    assert len(fonts) == 1


def test_create_page_with_locked_document(locked_document):
    with pytest.raises(LockedDocumentError):
        _ = locked_document.create_page(0)


def test_page_count_with_locked_document(locked_document):
    with pytest.raises(LockedDocumentError):
        _ = locked_document.pages


def test_create_font_iterator_with_locked_document(locked_document):
    with pytest.raises(LockedDocumentError):
        _ = locked_document.create_font_iterator()


def test_create_toc_with_locked_document(locked_document):
    with pytest.raises(LockedDocumentError):
        _ = locked_document.create_toc()


@pytest.mark.parametrize(
    "prop",
    (
        "author",
        "creation_date",
        "creator",
        "keywords",
        "metadata",
        "modification_date",
        "producer",
        "subject",
        "title",
        "page_layout",
        "page_mode",
    ),
)
def test_get_property_with_locked_document(prop, locked_document):
    with pytest.raises(LockedDocumentError):
        _ = getattr(locked_document, prop).fget()


@pytest.mark.parametrize(
    "prop",
    (
        "author",
        "creation_date",
        "creator",
        "keywords",
        "modification_date",
        "producer",
        "subject",
        "title",
    ),
)
def test_set_property_with_locked_document(prop, locked_document):
    with pytest.raises(LockedDocumentError):
        _ = getattr(locked_document, prop).fset("foobar")


def test_pdf_id_of_locked_document(locked_document):
    pdf_id = locked_document.pdf_id
    assert pdf_id.permanent_id == "c2a52b07e1fd69cc3d4fed8a2ea67ab6"
    assert pdf_id.update_id == "c2a52b07e1fd69cc3d4fed8a2ea67ab6"


def test_get_pdf_version_of_locked_document(locked_document):
    with pytest.raises(LockedDocumentError):
        _ = locked_document.pdf_version


@pytest.mark.skipif(version() < (0, 74, 0), reason="Requires at least Poppler 0.74.0")
def test_create_destination_map_of_locked_document(locked_document):
    # This one do not crash, but it doesn't make sense to query for locked document
    with pytest.raises(LockedDocumentError):
        _ = locked_document.create_destination_map()


def test_embedded_files_of_locked_document(locked_document):
    # This one do not crash, but it doesn't make sense to query for locked document
    with pytest.raises(LockedDocumentError):
        _ = locked_document.embedded_files()


def test_fonts_of_locked_document(locked_document):
    with pytest.raises(LockedDocumentError):
        _ = locked_document.fonts()


def test_locked_document_has_embedded_files(locked_document):
    with pytest.raises(LockedDocumentError):
        _ = locked_document.has_embedded_files()


@pytest.mark.parametrize(
    "permission",
    [
        Permission.print,
        Permission.change,
        Permission.copy,
        Permission.add_notes,
        Permission.fill_forms,
        Permission.accessibility,
        Permission.assemble,
        Permission.print_high_resolution,
    ],
)
def test_locked_document_permission(permission, locked_document):
    assert locked_document.has_permission(permission)


def test_locked_document_is_encrypted(locked_document):
    # not sure why it returns False when the document is locked...
    assert not locked_document.is_encrypted()


def test_locked_document_is_linearized(locked_document):
    assert not locked_document.is_linearized()



# poppler-python: python binding to the poppler-cpp pdf lib
# Copyright (C) 2020, Charles Brunet <charles@cbrunet.net>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from poppler.cpp import page_renderer
from poppler.cpp.global_ import rotation_enum as Rotation
from poppler.image import Image
from poppler.utilities import version, since

if version() >= (0, 65, 0):
    LineMode = page_renderer.line_mode_enum

RenderHint = page_renderer.render_hint
"""A flag of an option taken into account when rendering"""


class PageRenderer:
    """:class:`PageRenderer` allows to render a :class:`.Page` object
    to an :class:`.Image`.

    It is a wrapper for :class:`poppler.cpp.page_renderer.page_renderer`.

    """

    def __init__(self):
        self._renderer = page_renderer.page_renderer()

    @property
    @since(0, 65)
    def image_format(self):
        return self._renderer.image_format()

    @image_format.setter
    @since(0, 65)
    def image_format(self, format):
        self._renderer.set_image_format(format)

    @property
    @since(0, 65)
    def line_mode(self):
        return self._renderer.line_mode()

    @line_mode.setter
    @since(0, 65)
    def line_mode(self, mode):
        self._renderer.set_line_mode(mode)

    @property
    def paper_color(self):
        return self._renderer.paper_color()

    @paper_color.setter
    def paper_color(self, color):
        self._renderer.set_paper_color(color)

    @property
    def render_hints(self):
        return self._renderer.render_hints()

    @render_hints.setter
    def render_hints(self, hints):
        self._renderer.set_render_hints(hints)

    def render_page(
        self,
        page,
        xres=72.0,
        yres=72.0,
        x=-1,
        y=-1,
        w=-1,
        h=-1,
        rotate=Rotation.rotate_0,
    ):
        img = self._renderer.render_page(page._page, xres, yres, x, y, w, h, rotate)
        return Image.from_object(img)

    def set_render_hint(self, hint, on=True):
        self._renderer.set_render_hint(hint, on)

    @staticmethod
    def can_render():
        """Tell whether poppler was compiled with the Splash render backend.

        This should always return True. If it returns False,
        :meth:`.render_page` will systematically return
        an invalid :class:`.Image`.

        Returns:
            bool: whether a render backend was compiled with poppler.

        """
        return page_renderer.can_render()



# poppler-python: python binding to the poppler-cpp pdf lib
# Copyright (C) 2020, Charles Brunet <charles@cbrunet.net>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from poppler.cpp import rectangle


class Rectangle:
    def __init__(self, x=0, y=0, w=0, h=0, poppler_object=None):
        if poppler_object:
            self._rect = poppler_object
        elif all(map(lambda i: isinstance(i, int), (x, y, w, h))):
            self._rect = rectangle.rect(int(x), int(y), int(w), int(h))
        else:
            self._rect = rectangle.rectf(float(x), float(y), float(w), float(h))

    @classmethod
    def from_object(cls, poppler_object):
        return cls(poppler_object=poppler_object)

    @property
    def x(self):
        return self._rect.x()

    @property
    def y(self):
        return self._rect.y()

    def is_empty(self):
        return self._rect.is_empty()

    @property
    def height(self):
        return self._rect.height()

    @property
    def width(self):
        return self._rect.width()

    @property
    def bottom(self):
        return self._rect.bottom()

    @bottom.setter
    def bottom(self, value):
        self._rect.set_bottom(value)

    @property
    def left(self):
        return self._rect.left()

    @left.setter
    def left(self, value):
        self._rect.set_left(value)

    @property
    def right(self):
        return self._rect.right()

    @right.setter
    def right(self, value):
        self._rect.set_right(value)

    @property
    def top(self):
        return self._rect.top()

    @top.setter
    def top(self, value):
        self._rect.set_top(value)

    def as_tuple(self):
        return (self.left, self.top, self.right, self.bottom)



# poppler-python: python binding to the poppler-cpp pdf lib
# Copyright (C) 2020, Charles Brunet <charles@cbrunet.net>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from poppler.cpp import page, global_
from poppler.cpp.global_ import rotation_enum as Rotation
from poppler.pagetransition import PageTransition
from poppler.rectangle import Rectangle
from poppler.utilities import since, version


class _MetaTextBox(type):
    def __new__(mcs, classname, bases, dictionary):
        if version() >= (0, 89, 0):
            dictionary["WritingMode"] = page.writing_mode_enum
        return type.__new__(mcs, classname, bases, dictionary)


@since(0, 63)
class TextBox(metaclass=_MetaTextBox):
    def __init__(self, text_box):
        self._text_box = text_box

    @property
    def text(self):
        return str(self._text_box.text())

    @property
    def bbox(self):
        return Rectangle.from_object(self._text_box.bbox())

    @property
    @since(0, 68)
    def rotation(self):
        return self._text_box.rotation()

    def char_bbox(self, i):
        return Rectangle.from_object(self._text_box.char_bbox(i))

    @property
    def has_space_after(self):
        return self._text_box.has_space_after()

    @property
    @since(0, 89)
    def has_font_info(self):
        return self._text_box.has_font_info()

    @since(0, 89)
    def get_font_name(self, i=0):
        return self._text_box.get_font_name(i)

    @since(0, 89)
    def get_font_size(self):
        return self._text_box.get_font_size()

    @since(0, 89)
    def get_wmode(self, i=0):
        return self._text_box.get_wmode(i)


class _MetaPage(type):
    def __new__(mcs, classname, bases, dictionary):
        if version() >= (0, 89, 0):
            dictionary["TextListOption"] = page.text_list_option_enum
        return type.__new__(mcs, classname, bases, dictionary)


class Page(metaclass=_MetaPage):

    Orientation = page.orientation_enum
    PageBox = global_.page_box_enum
    SearchDirection = page.search_direction_enum
    TextLayout = page.text_layout_enum

    def __init__(self, poppler_page):
        self._page = poppler_page

    @property
    def duration(self):
        return self._page.duration()

    @property
    def label(self):
        return str(self._page.label())

    @property
    def orientation(self):
        return self._page.orientation()

    def page_rect(self, box=PageBox.crop_box):
        rectf = self._page.page_rect(box)
        return Rectangle.from_object(rectf)

    def search(self, text, r, direction, case_sensitivity, rotation=Rotation.rotate_0):
        utext = global_.ustring(text)
        found, found_rect = self._page.search(
            utext, r._rect, direction, case_sensitivity, rotation
        )
        if found:
            return Rectangle.from_object(found_rect)
        return None

    def text(self, rect=None, layout_mode=None):
        r = rect or Rectangle(0.0, 0.0, 0.0, 0.0)
        if layout_mode is None:
            t = self._page.text(r._rect)
        else:
            t = self._page.text(r._rect, layout_mode)
        return str(t)

    @since(0, 63)
    def text_list(self, opt_flag=None):
        if opt_flag is not None and version() >= (0, 89, 0):
            boxes = self._page.text_list(opt_flag)
        else:
            boxes = self._page.text_list()
        return [TextBox(b) for b in boxes]

    def transition(self):
        transition = self._page.transition()
        return PageTransition(transition) if transition else None
